@@ -122,31 +122,41 @@ sub _dumper {
 
 sub _dump_global {
     my ($item, $seen, $dumper_seen, $vars, $tab) = @_;
-    my $n = $item->{sigil} . $item->{namespace} . "::" . $item->{name};
-    if (!$seen->{$n}) {
-        if ($item->{sigil} eq '$') {
-            push @$vars, $tab . "$n = " . _dumper( eval $n, "  ", $dumper_seen, $n ) . ";\n";
+
+    if (ref($item) eq 'Perlito5::AST::Sub') {
+        my $n = $item->{namespace} . "::" . $item->{name};
+        if (!$seen->{$n}) {
+            push @$vars, $tab . "sub $n = " . _dumper( $item, "  ", $dumper_seen, $n ) . ";\n";
+            $seen->{$n} = 1;
         }
-        elsif ($item->{sigil} eq '@' || $item->{sigil} eq '%') {
-            my $ref = "\\$n";
-            my $d = _dumper( eval $ref, $tab . "  ", $dumper_seen, $ref );
-            if ($d eq '[]' || $d eq '{}') {
-                push @$vars, $tab . "$n = ();\n"
+    }
+    elsif (ref($item) eq 'Perlito5::AST::Var') {
+        my $n = $item->{sigil} . $item->{namespace} . "::" . $item->{name};
+        if (!$seen->{$n}) {
+            if ($item->{sigil} eq '$') {
+                push @$vars, $tab . "$n = " . _dumper( eval $n, "  ", $dumper_seen, $n ) . ";\n";
             }
-            else {
-                push @$vars, $tab . "$n = " . $item->{sigil} . "{" . $d . "};\n";
+            elsif ($item->{sigil} eq '@' || $item->{sigil} eq '%') {
+                my $ref = "\\$n";
+                my $d = _dumper( eval $ref, $tab . "  ", $dumper_seen, $ref );
+                if ($d eq '[]' || $d eq '{}') {
+                    push @$vars, $tab . "$n = ();\n"
+                }
+                else {
+                    push @$vars, $tab . "$n = " . $item->{sigil} . "{" . $d . "};\n";
+                }
             }
+            elsif ($item->{sigil} eq '*') {
+                # TODO - look for aliasing
+                #   *v1 = \$v2
+                push @$vars, $tab . "# $n\n";
+                for (qw/ $ @ % /) {
+                    local $item->{sigil} = $_;
+                    _dump_global($item, $seen, $dumper_seen, $vars, $tab);
+                }
+            }
+            $seen->{$n} = 1;
         }
-        elsif ($item->{sigil} eq '*') {
-            # TODO - look for aliasing
-            #   *v1 = \$v2
-            push @$vars, $tab . "# $n\n";
-            for (qw/ $ @ % /) {
-                local $item->{sigil} = $_;
-                _dump_global($item, $seen, $dumper_seen, $vars, $tab);
-            }
-        }
-        $seen->{$n} = 1;
     }
 }
 
@@ -165,7 +175,7 @@ sub _emit_globals {
         if (ref($item) eq 'Perlito5::AST::Var' && $item->{_decl} eq 'my') {
             my $id = $item->{_id};
             if (!$seen->{$id}) {
-                push @$vars, $tab . "# my " . $item->{sigil} . $item->{name} . ";\n";
+                push @$vars, $tab . emit_compiletime_lexical($item) . "  # my " . $item->{sigil} . $item->{name} . "\n";
             }
             $seen->{$id} = 1;
         }
@@ -178,7 +188,13 @@ sub _emit_globals {
     }
 }
 
-sub emit_globals {
+sub emit_compiletime_lexical {
+    my $item = shift;
+    # the internal compile-time namespace is "C_"
+    return $item->{sigil} . 'C_::' . $item->{name} . "_" . $item->{_id};
+}
+
+sub emit_globals_scope {
     # return a structure with the global variable declarations
     # this is used to initialize the ahead-of-time program
     my $scope = shift() // $Perlito5::BASE_SCOPE;
@@ -190,5 +206,40 @@ sub emit_globals {
     return join("", @vars);
 }
 
+sub emit_globals {
+    # return a structure with the global variable declarations
+    # this is used to initialize the ahead-of-time program
+    my $scope = shift() // $Perlito5::GLOBAL;
+    my $vars = [];
+    my $seen = {};
+    my $dumper_seen = {};
+    my $tab = "";
+
+    # TODO
+    #   - move global variables from SCOPE to GLOBAL
+
+    # problems to look for:
+    #   - lexical variables shared across closures
+    #   - our variables
+    #   - in order to conserve memory, create subroutine stubs that expand into instrumented code
+    #       when called
+    #   - bootstrapping rewrites Perlito5::* subroutines and globals, probably breaking the compiler
+
+    # example of how to enable this dump:
+    #   $ PERLITO5DEV=1 perl perlito5.pl -Isrc5/lib -I. -It -C_globals -e ' use X; xxx(); sub xyz { 123 } my $z; BEGIN { $a = 3; $z = 3 } '
+
+    for my $name (keys %$scope) {
+        my $item = $scope->{$name};
+        if (ref($item) eq 'Perlito5::AST::Sub' && $item->{name}) {
+            _dump_global($item, $seen, $dumper_seen, $vars, $tab);
+        }
+        else {
+            $item->{value} = eval($name);
+            push @$vars, "$name = " . _dumper( $item, "  ", $dumper_seen, $name . "->{value}" ) . ";\n";
+        }
+    }
+    return join("", @$vars);
+}
+
 1;
 
